/*
 * Copyright (c) 2014, Stéphane Zuckerman
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UTILS_H
#define UTILS_H

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>

static inline void fatal(const char* msg) {
    if (errno != 0) {
        perror(msg);
        errno = 1;
    } else {
        fputs(msg,stderr);
    }
    exit(errno);
}

static inline void warn(const char* msg) {
    if (errno != 0) {
        perror(msg);
    } else {
        fputs(msg,stderr);
    }
}

static inline void* smalloc(size_t sz) {
    void *p = malloc(sz);
    if (!p) fatal("malloc");
    return p;
}

static inline void* scalloc(size_t nmemb, size_t size)
{
    void* p = calloc(nmemb,size);
    if (!p) fatal("calloc");
    return p;
}

static inline void* srealloc(void* p, size_t sz) {
    void* ptr = realloc(p,sz);
    if (!ptr) { 
        free(p); 
        fatal("realloc"); 
    }
    return ptr;
}

static inline void sfree(void* p) {
    *(char*)p = 0;
    free(p);
}

extern char* progname; // defined in the file with main(). Usually, progname == argv[0]
static inline void usage(const char* msg) {
    fprintf(stderr,"USAGE:\t%s %s\n", progname, msg ? msg : "");
    exit(EXIT_FAILURE);
}

static inline long read_int(const char* strnum) {
    const char *errmsg = "Invalid input: Not a valid number";
    char*  strerr = NULL;
    size_t n      = strtoul(strnum, &strerr,0);
    if ( n == 0 && strerr != '\0' )  {
        fprintf(stderr, "%s: %s\n", errmsg, strerr);
        exit(EXIT_FAILURE);
    }
    return n;
}

#endif // UTILS_H
