/*
 * Copyright (c) 2014, Stéphane Zuckerman
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "utils.h"
#include <stdbool.h>
#include <string.h>
#include <limits.h>
#include "linked_list.h"

void 
list_init 
(
     linked_list_t* list, 
     link_t*  (*new_link)     (int),
     void     (*destroy_link) (link_t*, int*),
     void     (*push_back)    (linked_list_t*,                      int),
     void     (*push_front)   (linked_list_t*,                      int),
     link_t*  (*find)         (linked_list_t*, link_t* start, const int) 
)
{
    *list = (linked_list_t){
        .new_link     = new_link,
        .destroy_link = destroy_link,
        .push_back    = push_back,
        .push_front   = push_front,
        .find         = find,
        .head         = NULL,
        .tail         = NULL,
        .size         = 0
    };
}

link_t* 
link_init(int value)
{
    link_t* link = scalloc(1,sizeof(link_t));
    link->data   = value;
    link->next   = NULL;
    return link;
}

void
link_destroy(link_t* link, int* value)
{
    *value = INT_MIN;
    if (link != NULL) {
        *value = link->data;
        free(link);
    }
}

void 
list_destroy (struct linked_list_s* list)
{
    int retval = INT_MIN;
    while (list->head->next != NULL)
        remove_after(list,NULL,&retval);
//  while (list->head != NULL) {
//      link_t *cur = list->head;
//      list->head  = cur->next;
//      memset(cur,0,sizeof(link_t));
//      free(cur);
//  }
    memset(list,0,sizeof(linked_list_t));
}

void 
push_back (struct linked_list_s* list, int value) 
{
    link_t* link = list->new_link(value);

    if (list->size++ == 0) { // Empty list
        list->head = link;
        list->tail = link;
    } else {
        list->tail->next = link; // chain link to current tail 
        list->tail       = link; // update tail link
    }
}

void 
push_front (struct linked_list_s* list, int value)
{
    link_t* link = list->new_link(value);
    
    if (list->size++ == 0) { // Empty list
        list->head = link;
        list->tail = link;
    } else {
        link->next = list->head;
        list->head = link;
    }
}

link_t*
find (struct linked_list_s* list, link_t* start, const int value) 
{
    link_t* cur = start ? start : list->head;
    for (; cur != NULL && cur->data != value; cur = cur->next) 
        ;
    return cur;
}

void
insert_after(linked_list_t* list, link_t* l, int value)
{
    link_t* link = list->new_link(value);

    if (list->size++ == 0) { // empty list
        list->head = link;
        list->tail = link;
    } else {
        if (l == NULL) { // insert at head
            link->next       = list->head;
            list->head       = link;
        } else {
            link->next = l;
            l->next    = link;
        }
    }
}

void
remove_after(linked_list_t* list, link_t* l, int* value)
{
    l            = l == NULL ? list->head : l;
    link_t* link = l == NULL ?       NULL : l->next; 

    if (link != NULL) {
        list->tail = l->next == list->tail ? l : list->tail;
        *value     = link->data;
        l->next    = link->next;
        memset(link,0,sizeof(link_t));
        free(link);
        if (--list->size == 0)
            list->tail = NULL;
    }
}
