/*
 * Copyright (c) 2014, Stéphane Zuckerman
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "utils.h"
#include <stdbool.h>
#include <string.h>
#include <limits.h>
#include "linked_list_pthd.h"

void 
pthd_list_init 
(
     pthd_linked_list_t* list, 
     pthd_link_t*  (*new_link)     (pthread_t),
     void          (*destroy_link) (pthd_link_t*, pthread_t*),
     void          (*push_back)    (pthd_linked_list_t*, pthread_t),
     void          (*push_front)   (pthd_linked_list_t*,                           pthread_t),
     pthd_link_t*  (*find)         (pthd_linked_list_t*, pthd_link_t* start, const pthread_t) 
)
{
    *list = (pthd_linked_list_t){
        .new_link     = new_link,
        .destroy_link = destroy_link,
        .push_back    = push_back,
        .push_front   = push_front,
        .find         = find,
        .head         = NULL,
        .tail         = NULL,
        .size         = 0
    };
}

pthd_link_t* 
pthd_link_init(pthread_t value)
{
    pthd_link_t* link = scalloc(1,sizeof(pthd_link_t));
    link->id          = value;
    link->next        = NULL;
    return link;
}

void
pthd_link_destroy(pthd_link_t* link, pthread_t* value)
{
    *value = INT_MIN;
    if (link != NULL) {
        *value = link->id;
        free(link);
    }
}

void 
pthd_list_destroy (struct pthd_linked_list_s* list)
{
    pthread_t retval;
    while (list->head->next != NULL)
        pthd_remove_after(list,NULL,&retval);
//  while (list->head != NULL) {
//      pthd_link_t *cur = list->head;
//      list->head  = cur->next;
//      memset(cur,0,sizeof(pthd_link_t));
//      free(cur);
//  }
    memset(list,0,sizeof(pthd_linked_list_t));
}

void 
pthd_push_back (struct pthd_linked_list_s* list, pthread_t value) 
{
    pthd_link_t* link = list->new_link(value);

    if (list->size++ == 0) { // Empty list
        list->head = link;
        list->tail = link;
    } else {
        list->tail->next = link; // chain link to current tail 
        list->tail       = link; // update tail link
    }
}

void 
pthd_push_front (struct pthd_linked_list_s* list, pthread_t value)
{
    pthd_link_t* link = list->new_link(value);
    
    if (list->size++ == 0) { // Empty list
        list->head = link;
        list->tail = link;
    } else {
        link->next = list->head;
        list->head = link;
    }
}

pthd_link_t*
pthd_find (struct pthd_linked_list_s* list, pthd_link_t* start, const pthread_t value) 
{
    pthd_link_t* cur = start ? start : list->head;
    for (; cur != NULL && cur->id != value; cur = cur->next) 
        ;
    return cur;
}

void
pthd_insert_after(pthd_linked_list_t* list, pthd_link_t* l, pthread_t value)
{
    pthd_link_t* link = list->new_link(value);

    if (list->size++ == 0) { // empty list
        list->head = link;
        list->tail = link;
    } else {
        if (l == NULL) { // insert at head
            link->next       = list->head;
            list->head       = link;
        } else {
            link->next = l;
            l->next    = link;
        }
    }
}

void
pthd_remove_after(pthd_linked_list_t* list, pthd_link_t* l, pthread_t* value)
{
    l            = l == NULL ? list->head : l;
    pthd_link_t* link = l == NULL ?       NULL : l->next; 

    if (link != NULL) {
        list->tail = l->next == list->tail ? l : list->tail;
        *value     = link->id;
        l->next    = link->next;
        memset(link,0,sizeof(pthd_link_t));
        free(link);
        if (--list->size == 0)
            list->tail = NULL;
    }
}
