#include "utils.h"
#include "pthread_wrappers.h"
#include <stddef.h>

const size_t N_THREADS = 4;
const double EPSILON   = 0.0001;

const size_t M = 10;
const size_t N = 10;

typedef struct matvec_s {
    uint64_t      id,
                  n_threads;
    double       *v; 
    const double *A, 
                 *b;
    size_t        m, 
                  n;
} matvec_t;

void*
thd_matvec_task(void* arg)
{
    matvec_t* frame = (matvec_t*) arg;
    const double *A = frame->A, 
                 *b = frame->b;
    double       *v = frame->v;
            
    uint64_t  lo    =    frame->id  * M / frame->n_threads,
              hi    = (1+frame->id) * M / frame->n_threads;

    for (size_t i = lo; i < hi; ++i) {
        double acc = 0.0;
        for (size_t j = 0; j < frame->n; ++j) {
            acc += A[i*frame->n+j] * b[j];
        }
        v[i] = acc;
    }

    return NULL;
}

void
thd_matvec(double* v, const double* A, const double* b, size_t m, size_t n)
{
    pthread_t      tasks [N_THREADS];
    matvec_t       args  [N_THREADS];
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    /* pedantic: default value is already PTHREAD_CREATE_JOINABLE */
    pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_JOINABLE); 

    for (size_t i = 0; i < N_THREADS; ++i) {
        args[i] = (matvec_t) {
          .id = i, .n_threads = N_THREADS, .v = v, .A = A, .b = b, .m = m, .n = n
        };
        spthread_create(tasks+i,&attr,thd_matvec_task,args+i);
    }

    for (size_t i = 0; i < N_THREADS; ++i) 
        spthread_join(tasks[i],NULL);
}

void
seq_matvec(double* v, const double* A, const double* b, size_t m, size_t n)
{
    for (size_t i = 0; i < m; ++i) {
        double acc = 0.0;
        for (size_t j = 0; j < n; ++j) {
            acc += A[i*n+j] * b[j];
        }
        v[i] = acc;
    }
}


int
main(void)
{
    double *A = smalloc(M * N * sizeof(double)),
           *b = smalloc(    N * sizeof(double)),
           *v = scalloc(N,      sizeof(double));

    for (size_t i = 0; i < M; ++i) 
        for (size_t j = 0; j < N; ++j) 
            A[i*N+j] = i + j*2.01;

    for (size_t i = 0; i < M; ++i) 
        b[i] = 10.3 * i + 1;

    thd_matvec(v,A,b,M,N);

    // Sequential run
    double *v_seq = scalloc(N,      sizeof(double));
    seq_matvec(v_seq,A,b,M,N);

int errors=0;
    for (size_t i = 0; i < M; ++i) {
        if (v_seq[i]-v[i] > EPSILON || v_seq[i]-v[i] < -EPSILON)
{
            printf("v_seq[%lu] = %.4f != v[%lu] = %.4f\n", i, v_seq[i], i, v[i]);
errors++;
}
    }
if(errors!=0)
{
printf("%d errors found!\n", errors);
}
else
{
printf("No errors found!\n");
}



    sfree(v_seq);
    sfree(v);
    sfree(b);
    sfree(A);
}
