/*
 * Copyright (c) 2014, Stéphane Zuckerman
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef PTHREAD_WRAPPERS_H_GUARD
#define PTHREAD_WRAPPERS_H_GUARD

#include <pthread.h>
#include <assert.h>

#define ASSERT(condition,err_msg) assert((condition) && (err_msg))
#define FATAL_ERR(msg) fprintf(stderr,"%s\n",(msg)), exit(1)
#define WARN(msg) fprintf(stderr, "%s", (msg))

static inline void 
spthread_create(pthread_t* thread, const pthread_attr_t* attr, 
                void* (*code)(void*), void* data)
{
    int err = pthread_create(thread,attr,code,data);
    switch ( err ) {
    case EAGAIN:
        WARN("pthread_create\t");
        FATAL_ERR("Insufficient  resources  to create another thread, or a "
                  "system-imposed limit on the number of threads was encountered.\n");
    case EINVAL: 
        WARN("pthread_create\t");
        FATAL_ERR("Invalid settings in attribute.\n");
    case EPERM:
        WARN("pthread_create\t");
        FATAL_ERR("No permission to set the scheduling policy and parameters"
                  " specified in attribute.\n");
    case 0:  break;
    default: break;
    }
}

static inline void 
spthread_join(pthread_t thread, void** retval)
{
    int err = pthread_join(thread,retval);
    switch ( err ) {
    case EDEADLK: 
        WARN("pthread_join\t");
        WARN("A deadlock was detected or thread tried to join itself\n");
        break;
    case EINVAL: 
        WARN("pthread_join\t");
        WARN("Thread is not a joinable thread OR Another thread is"
             " already waiting to join with this thread.\n");
        break;
    case ESRCH: 
        WARN("pthread_join\t");
        WARN("No thread with the ID thread could be found.\n");
        break;
    case 0:  break;
    default: break;
    }
}

static inline void spthread_mutex_lock(pthread_mutex_t* mutex)
{
    int err = pthread_mutex_lock(mutex);
    switch ( err ) {
    case EINVAL:  
        WARN("pthread_mutex_lock\t");
        FATAL_ERR("The mutex was created with the protocol attribute having"
                  " the value PTHREAD_PRIO_PROTECT and the calling thread's"
                  " priority is higher than the mutex's current priority ceiling."
                  " OR: The value specified by mutex does not refer to an"
                  " initialized mutex object.\n");
    case EAGAIN:  
        WARN("pthread_mutex_lock\t");
        WARN("The mutex could not be acquired because the maximum number of "
             "recursive locks for mutex has been exceeded.\n");
        FATAL_ERR("The value specified by mutex does not refer to an initialized "
                  "mutex object.\n");
    case EDEADLK: 
        WARN("pthread_mutex_lock\t");
        FATAL_ERR("The current thread already owns the mutex.\n");
    case 0:  break;
    default: break;
    }
}

static inline void
spthread_mutex_unlock(pthread_mutex_t* lock)
{
    int err = pthread_mutex_unlock(lock);
    switch ( err ) {
    case EAGAIN:  
        WARN("pthread_mutex_unlock\t");
        WARN("The mutex could not be acquired because the maximum number of "
             "recursive locks for mutex has been exceeded.\n");
        FATAL_ERR("The value specified by mutex does not refer to "
                  "an initialized mutex object.\n");
    case EINVAL:  
        WARN("pthread_mutex_unlock\t");
        FATAL_ERR("The value specified by mutex does not refer to "
                  "an initialized mutex object.\n");
    case EDEADLK: 
        WARN("pthread_mutex_unlock\t");
        FATAL_ERR("The current thread already owns the mutex.\n");
    case 0:  break;
    default: break;
    }
}

static inline void 
spthread_cond_wait(pthread_cond_t*  restrict cond, 
                   pthread_mutex_t* restrict mutex)
{
    int err = pthread_cond_wait(cond, mutex);
    ASSERT( !err, "pthread_cond_wait failed" );
}

static inline void
spthread_cond_broadcast(pthread_cond_t *cond)
{
    int err = pthread_cond_broadcast(cond);
    ASSERT( !err, "pthread_cond_wait failed" );
}

#endif
